"""Python library to interact with UniFi controller."""

from __future__ import annotations

from collections.abc import Callable
from http import HTTPStatus
import logging
from pprint import pformat
from ssl import SSLContext
from typing import TYPE_CHECKING, Any

import aiohttp
from aiohttp import client_exceptions

from .errors import (
    BadGateway,
    Forbidden,
    LoginRequired,
    RequestError,
    ResponseError,
    ServiceUnavailable,
    raise_error,
)
from .interfaces.clients import Clients
from .interfaces.clients_all import ClientsAll
from .interfaces.devices import Devices
from .interfaces.dpi_restriction_apps import DPIRestrictionApps
from .interfaces.dpi_restriction_groups import DPIRestrictionGroups
from .interfaces.events import EventHandler
from .interfaces.messages import MessageHandler
from .interfaces.outlets import Outlets
from .interfaces.ports import Ports
from .interfaces.wlans import Wlans
from .models.site import SiteDescriptionRequest, SiteListRequest
from .websocket import WebsocketSignal, WebsocketState, WSClient

if TYPE_CHECKING:
    from .models.request_object import RequestObject

LOGGER = logging.getLogger(__name__)


class Controller:
    """Control a UniFi controller."""

    def __init__(
        self,
        host: str,
        websession: aiohttp.ClientSession,
        *,
        username: str,
        password: str,
        port: int = 8443,
        site: str = "default",
        sslcontext: SSLContext | None = None,
        callback: Callable[[WebsocketSignal, dict[str, Any] | WebsocketState], None]
        | None = None,
    ) -> None:
        """Session setup."""
        self.host = host
        self.session = websession
        self.port = port
        self.username = username
        self.password = password
        self.site = site
        self.sslcontext = sslcontext
        self.callback = callback
        self.can_retry_login = False

        self.url = f"https://{self.host}:{self.port}"
        self.is_unifi_os = False
        self.headers: dict[str, Any] = {}
        self.last_response: aiohttp.ClientResponse | None = None

        self.websocket: WSClient | None = None

        self.messages = MessageHandler(self)
        self.events = EventHandler(self)

        self.clients = Clients(self)
        self.clients_all = ClientsAll(self)
        self.devices = Devices(self)
        self.outlets = Outlets(self)
        self.ports = Ports(self)
        self.dpi_apps = DPIRestrictionApps(self)
        self.dpi_groups = DPIRestrictionGroups(self)
        self.wlans = Wlans(self)

    async def check_unifi_os(self) -> None:
        """Check if controller is running UniFi OS."""
        await self._request("get", self.url, allow_redirects=False)
        if (
            response := self.last_response
        ) is not None and response.status == HTTPStatus.OK:
            self.is_unifi_os = True
        LOGGER.debug("Talking to UniFi OS device: %s", self.is_unifi_os)

    async def login(self) -> None:
        """Log in to controller."""
        if self.is_unifi_os:
            url = f"{self.url}/api/auth/login"
        else:
            url = f"{self.url}/api/login"

        auth = {
            "username": self.username,
            "password": self.password,
            "remember": True,
        }

        await self._request("post", url, json=auth)

        if (
            (response := self.last_response) is not None
            and response.status == HTTPStatus.OK
            and (csrf_token := response.headers.get("x-csrf-token")) is not None
        ):
            self.headers = {"x-csrf-token": csrf_token}

        self.can_retry_login = True

    async def sites(self) -> dict[str, Any]:
        """Retrieve what sites are provided by controller."""
        sites = await self.request(SiteListRequest.create())
        LOGGER.debug(pformat(sites))
        return {site["desc"]: site for site in sites}

    async def site_description(self) -> list[dict[str, Any]]:
        """User description of current site."""
        description = await self.request(SiteDescriptionRequest.create())
        LOGGER.debug(description)
        return description

    async def initialize(self) -> None:
        """Load UniFi parameters."""
        await self.clients.update()
        await self.clients_all.update()
        await self.devices.update()
        await self.dpi_apps.update()
        await self.dpi_groups.update()
        await self.wlans.update()

    def start_websocket(self) -> None:
        """Start websession and websocket to UniFi."""
        self.websocket = WSClient(
            self.session,
            self.host,
            self.port,
            self.sslcontext,
            self.site,
            callback=self.session_handler,
            is_unifi_os=self.is_unifi_os,
        )
        self.websocket.start()

    def stop_websocket(self) -> None:
        """Close websession and websocket to UniFi."""
        LOGGER.info("Shutting down connections to UniFi.")
        if self.websocket:
            self.websocket.stop()

    def session_handler(self, signal: WebsocketSignal) -> None:
        """Signalling from websocket.

        data - new data available for processing.
        state - network state has changed.
        """
        assert self.websocket

        if signal == WebsocketSignal.DATA:
            new_items = self.messages.handler(self.websocket.data)
            if new_items and self.callback:
                self.callback(WebsocketSignal.DATA, new_items)

        elif signal == WebsocketSignal.CONNECTION_STATE and self.callback:
            self.callback(WebsocketSignal.CONNECTION_STATE, self.websocket.state)

    async def request(self, request_object: RequestObject) -> list[dict[str, Any]]:
        """Make a request to the API, retry login on failure."""
        url = self.url + request_object.full_path(self.site, self.is_unifi_os)

        try:
            response: list[dict[str, Any]] = await self._request(
                request_object.method, url, request_object.data
            )
            return response

        except LoginRequired:
            if not self.can_retry_login:
                raise
            # Session likely expired, try again
            self.can_retry_login = False
            await self.login()
            return await self.request(request_object)

    async def _request(
        self,
        method: str,
        url: str,
        json: dict[str, Any] | None = None,
        **kwargs: bool,
    ) -> list[dict[str, Any]] | Any:
        """Make a request to the API."""
        self.last_response = None

        LOGGER.debug("%s", url)

        try:
            async with self.session.request(
                method,
                url,
                json=json,
                ssl=self.sslcontext,
                headers=self.headers,
                **kwargs,
            ) as res:
                LOGGER.debug("%s %s %s", res.status, res.content_type, res)

                self.last_response = res

                if res.status == HTTPStatus.UNAUTHORIZED:
                    raise LoginRequired(f"Call {url} received 401 Unauthorized")

                if res.status == HTTPStatus.FORBIDDEN:
                    raise Forbidden(f"Call {url} received 403 Forbidden")

                if res.status == HTTPStatus.NOT_FOUND:
                    raise ResponseError(f"Call {url} received 404 Not Found")

                if res.status == HTTPStatus.BAD_GATEWAY:
                    raise BadGateway(f"Call {url} received 502 bad gateway")

                if res.status == HTTPStatus.SERVICE_UNAVAILABLE:
                    raise ServiceUnavailable(
                        f"Call {url} received 503 service unavailable"
                    )

                if res.content_type == "application/json":
                    response = await res.json()
                    _raise_on_error(response)
                    if "data" in response:
                        return response["data"]
                    return response
                return []

        except client_exceptions.ClientError as err:
            raise RequestError(
                f"Error requesting data from {self.host}: {err}"
            ) from None


def _raise_on_error(data: dict[str, Any] | None) -> None:
    """Check response for error message."""
    if not isinstance(data, dict):
        return None

    if "meta" in data and data["meta"]["rc"] == "error":
        raise_error(data["meta"]["msg"])

    if "errors" in data:
        raise_error(data["errors"][0])



"""API management class and base class for the different end points."""

from __future__ import annotations

from collections.abc import Callable, ItemsView, Iterator, ValuesView
import enum
from typing import TYPE_CHECKING, Any, Generic, Optional, final

from ..models import ResourceType
from ..models.request_object import RequestObject

if TYPE_CHECKING:
    from ..controller import Controller
    from ..models.event import Event, EventKey
    from ..models.message import Message, MessageKey


class ItemEvent(enum.Enum):
    """The event action of the item."""

    ADDED = "added"
    CHANGED = "changed"
    DELETED = "deleted"


CallbackType = Callable[[ItemEvent, str], None]
SubscriptionType = tuple[CallbackType, Optional[tuple[ItemEvent, ...]]]
UnsubscribeType = Callable[[], None]

ID_FILTER_ALL = "*"


class SubscriptionHandler:
    """Manage subscription and notification to subscribers."""

    def __init__(self) -> None:
        """Initialize subscription handler."""
        self._subscribers: dict[str, list[SubscriptionType]] = {ID_FILTER_ALL: []}

    def signal_subscribers(self, event: ItemEvent, obj_id: str) -> None:
        """Signal subscribers."""
        subscribers: list[SubscriptionType] = (
            self._subscribers.get(obj_id, []) + self._subscribers[ID_FILTER_ALL]
        )
        for callback, event_filter in subscribers:
            if event_filter is not None and event not in event_filter:
                continue
            callback(event, obj_id)

    def subscribe(
        self,
        callback: CallbackType,
        event_filter: tuple[ItemEvent, ...] | ItemEvent | None = None,
        id_filter: tuple[str] | str | None = None,
    ) -> UnsubscribeType:
        """Subscribe to added events."""
        if isinstance(event_filter, ItemEvent):
            event_filter = (event_filter,)
        subscription = (callback, event_filter)

        _id_filter: tuple[str]
        if id_filter is None:
            _id_filter = (ID_FILTER_ALL,)
        elif isinstance(id_filter, str):
            _id_filter = (id_filter,)

        for obj_id in _id_filter:
            if obj_id not in self._subscribers:
                self._subscribers[obj_id] = []
            self._subscribers[obj_id].append(subscription)

        def unsubscribe() -> None:
            for obj_id in _id_filter:
                if obj_id not in self._subscribers:
                    continue
                if subscription not in self._subscribers[obj_id]:
                    continue
                self._subscribers[obj_id].remove(subscription)

        return unsubscribe


class APIHandler(SubscriptionHandler, Generic[ResourceType]):
    """Base class for a map of API Items."""

    obj_id_key: str
    path: str
    item_cls: Any
    events: tuple[EventKey, ...] = ()
    process_messages: tuple[MessageKey, ...] = ()
    remove_messages: tuple[MessageKey, ...] = ()

    def __init__(self, controller: Controller) -> None:
        """Initialize API handler."""
        super().__init__()
        self.controller = controller
        self._items: dict[int | str, ResourceType] = {}

        if message_filter := self.process_messages + self.remove_messages:
            controller.messages.subscribe(self.process_message, message_filter)

        if self.events:
            controller.events.subscribe(self.process_event, self.events)

    @final
    async def update(self) -> None:
        """Refresh data."""
        raw = await self.controller.request(RequestObject("get", self.path, None))
        self.process_raw(raw)

    def process_raw(self, raw: list[dict[str, Any]]) -> set[str]:
        """Process full raw response."""
        new_items: set[str] = set()
        for raw_item in raw:
            if obj_id := self.process_item(raw_item):
                new_items.add(obj_id)
        return new_items

    def process_message(self, message: Message) -> str:
        """Process and forward websocket data."""
        if message.meta.message in self.process_messages:
            return self.process_item(message.data)

        if message.meta.message in self.remove_messages:
            return self.remove_item(message.data)

        return ""

    @final
    def process_event(self, event: Event) -> None:
        """Process event."""
        if (obj := self._items.get(event.mac)) is not None:
            obj.update(event=event)

    @final
    def process_item(self, raw: dict[str, Any]) -> str:
        """Process item data."""
        obj_id: str

        if self.obj_id_key not in raw:
            return ""

        if (obj_id := raw[self.obj_id_key]) in self._items:
            obj = self._items[obj_id]
            obj.update(raw=raw)
            self.signal_subscribers(ItemEvent.CHANGED, obj_id)
            return ""

        self._items[obj_id] = self.item_cls(raw, self.controller)
        self.signal_subscribers(ItemEvent.ADDED, obj_id)

        return obj_id

    @final
    def remove_item(self, raw: dict[str, Any]) -> str:
        """Remove item."""
        obj_id: str
        if (obj_id := raw[self.obj_id_key]) in self._items:
            obj = self._items.pop(obj_id)
            obj.clear_callbacks()
            self.signal_subscribers(ItemEvent.DELETED, obj_id)
            return obj_id
        return ""

    @final
    def items(self) -> ItemsView[int | str, ResourceType]:
        """Return items dictionary."""
        return self._items.items()

    @final
    def values(self) -> ValuesView[ResourceType]:
        """Return items."""
        return self._items.values()

    @final
    def get(
        self,
        obj_id: int | str,
        default: Any | None = None,
    ) -> ResourceType | None:
        """Get item value based on key, return default if no match."""
        return self._items.get(obj_id, default)

    @final
    def __contains__(self, obj_id: int | str) -> bool:
        """Validate membership of item ID."""
        return obj_id in self._items

    @final
    def __getitem__(self, obj_id: int | str) -> ResourceType:
        """Get item value based on key."""
        return self._items[obj_id]

    @final
    def __iter__(self) -> Iterator[int | str]:
        """Allow iterate over items."""
        return iter(self._items)



"""Device port implementation."""

from __future__ import annotations

from dataclasses import dataclass

from .device import TypedDevicePortTable


@dataclass
class Port2:
    """Network port representation."""

    ifname: str | None
    media: str | None
    name: str
    port_idx: int | None
    poe_class: str | None
    poe_enable: bool | None
    poe_mode: str | None
    poe_power: str | None
    poe_voltage: str | None
    portconf_id: str | None
    port_poe: bool | None
    up: bool | None

    @classmethod
    def from_dict(cls, data: TypedDevicePortTable) -> "Port2":
        """Create data container instance from dict."""
        return cls(
            ifname=data.get("ifname"),  # Port index as name used by USG
            media=data.get("media"),  # Media port is connected to
            name=data["name"],  # Port name
            port_idx=data.get("port_idx"),  # Port index
            poe_class=data.get("poe_class"),  # Port PoE class
            poe_enable=data.get("poe_enable"),  # Is PoE supported/requested by client
            poe_mode=data.get("poe_mode"),  # auto, pasv24, passthrough, off or None
            poe_power=data.get("poe_power"),  # PoE power usage
            poe_voltage=data.get("poe_voltage"),  # PoE voltage usage
            portconf_id=data.get("portconf_id"),  # Port configuration ID
            port_poe=data.get("port_poe"),  # Is PoE used
            up=data.get("up"),  # Is port up
        )


class Port:
    """Represents a network port."""

    def __init__(self, raw: TypedDevicePortTable) -> None:
        """Initialize port."""
        self.raw = raw

    @property
    def ifname(self) -> str | None:
        """Port name used by USG."""
        return self.raw.get("ifname")

    @property
    def media(self) -> str | None:
        """Media port is connected to."""
        return self.raw.get("media")

    @property
    def name(self) -> str:
        """Port name."""
        return self.raw["name"]

    @property
    def port_idx(self) -> int | None:
        """Port index."""
        return self.raw.get("port_idx")

    @property
    def poe_class(self) -> str | None:
        """Port POE class."""
        return self.raw.get("poe_class")

    @property
    def poe_enable(self) -> bool | None:
        """Is POE supported/requested by client."""
        return self.raw.get("poe_enable")

    @property
    def poe_mode(self) -> str | None:
        """Is POE auto, pasv24, passthrough, off or None."""
        return self.raw.get("poe_mode")

    @property
    def poe_power(self) -> str | None:
        """POE power usage."""
        return self.raw.get("poe_power")

    @property
    def poe_voltage(self) -> str | None:
        """POE voltage usage."""
        return self.raw.get("poe_voltage")

    @property
    def portconf_id(self) -> str | None:
        """Port configuration ID."""
        return self.raw.get("portconf_id")

    @property
    def port_poe(self) -> bool | None:
        """Is POE used."""
        return self.raw.get("port_poe")

    @property
    def up(self) -> bool | None:
        """Is port up."""
        return self.raw.get("up")

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<{self.name}: Poe {self.poe_enable}>"



"""Clients are devices on a UniFi network."""

from __future__ import annotations

from typing import Final

from ..models.client import Client
from .api_handlers import APIHandler

URL_ALL: Final = "/rest/user"  # All known and configured clients


class ClientsAll(APIHandler[Client]):
    """Represents all client network devices."""

    obj_id_key = "mac"
    path = URL_ALL
    item_cls = Client



"""Site is a specific grouping in a UniFi network."""

from __future__ import annotations

from dataclasses import dataclass

from .request_object import RequestObject


@dataclass
class SiteListRequest(RequestObject):
    """Request object for site list."""

    @classmethod
    def create(cls) -> "SiteListRequest":
        """Create site list request."""
        return cls(method="get", path="", data=None)

    def full_path(self, site: str, is_unifi_os: bool) -> str:
        """Url to list sites is global for controller."""
        if is_unifi_os:
            return "/proxy/network/api/self/sites"
        return "/api/self/sites"


@dataclass
class SiteDescriptionRequest(RequestObject):
    """Request object for site description."""

    @classmethod
    def create(cls) -> "SiteDescriptionRequest":
        """Create site list request."""
        return cls(method="get", path="/self", data=None)



"""Clients are devices on a UniFi network."""

from __future__ import annotations

from typing import Any, Final

from ..models.client import (
    Client,
    ClientBlockRequest,
    ClientReconnectRequest,
    ClientRemoveRequest,
)
from ..models.event import EventKey
from ..models.message import MessageKey
from .api_handlers import APIHandler

URL: Final = "/stat/sta"


class Clients(APIHandler[Client]):
    """Represents client network devices."""

    obj_id_key = "mac"
    path = URL
    item_cls = Client
    events = (
        EventKey.WIRED_CLIENT_CONNECTED,
        EventKey.WIRED_CLIENT_DISCONNECTED,
        EventKey.WIRED_CLIENT_BLOCKED,
        EventKey.WIRED_CLIENT_UNBLOCKED,
        EventKey.WIRELESS_CLIENT_CONNECTED,
        EventKey.WIRELESS_CLIENT_DISCONNECTED,
        EventKey.WIRELESS_CLIENT_BLOCKED,
        EventKey.WIRELESS_CLIENT_UNBLOCKED,
        EventKey.WIRELESS_CLIENT_ROAM,
        EventKey.WIRELESS_CLIENT_ROAMRADIO,
        EventKey.WIRELESS_GUEST_CONNECTED,
        EventKey.WIRELESS_GUEST_DISCONNECTED,
        EventKey.WIRELESS_GUEST_ROAM,
        EventKey.WIRELESS_GUEST_ROAMRADIO,
    )
    process_messages = (MessageKey.CLIENT,)
    remove_messages = (MessageKey.CLIENT_REMOVED,)

    async def block(self, mac: str) -> list[dict[str, Any]]:
        """Block client from controller."""
        return await self.controller.request(ClientBlockRequest.create(mac, block=True))

    async def unblock(self, mac: str) -> list[dict[str, Any]]:
        """Unblock client from controller."""
        return await self.controller.request(
            ClientBlockRequest.create(mac, block=False)
        )

    async def reconnect(self, mac: str) -> list[dict[str, Any]]:
        """Force a wireless client to reconnect to the network."""
        return await self.controller.request(ClientReconnectRequest.create(mac))

    async def remove_clients(self, macs: list[str]) -> list[dict[str, Any]]:
        """Make controller forget provided clients."""
        return await self.controller.request(ClientRemoveRequest.create(macs))



"""Data class for doing UniFi requests."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class RequestObject:
    """Data class with required properties of a request."""

    method: str
    path: str
    data: dict[str, Any] | None

    def full_path(self, site: str, is_unifi_os: bool) -> str:
        """Create url to work with a specific controller."""
        if is_unifi_os:
            return f"/proxy/network/api/s/{site}{self.path}"
        return f"/api/s/{site}{self.path}"



"""WLANs as part of a UniFi network."""

from typing import Any, Final

from ..models.wlan import Wlan, WlanEnableRequest
from .api_handlers import APIHandler

URL: Final = "/rest/wlanconf"  # List WLAN configuration


class Wlans(APIHandler[Wlan]):
    """Represents WLAN configurations."""

    obj_id_key = "name"
    path = URL
    item_cls = Wlan

    async def enable(self, wlan: Wlan) -> list[dict[str, Any]]:
        """Block client from controller."""
        return await self.controller.request(
            WlanEnableRequest.create(wlan.id, enable=True)
        )

    async def disable(self, wlan: Wlan) -> list[dict[str, Any]]:
        """Unblock client from controller."""
        return await self.controller.request(
            WlanEnableRequest.create(wlan.id, enable=False)
        )



"""DPI Restrictions as part of a UniFi network."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict

from .api import APIItem
from .request_object import RequestObject


class TypedDPIRestrictionApp(TypedDict):
    """DPI restriction app type definition."""

    _id: str
    apps: list[str]
    blocked: bool
    cats: list[str]
    enabled: bool
    log: bool
    site_id: str


@dataclass
class DPIRestrictionAppEnableRequest(RequestObject):
    """Request object for enabling DPI Restriction App."""

    @classmethod
    def create(cls, app_id: str, enable: bool) -> "DPIRestrictionAppEnableRequest":
        """Create enabling DPI Restriction App request."""
        return cls(
            method="put",
            path=f"/rest/dpiapp/{app_id}",
            data={"enabled": enable},
        )


class DPIRestrictionApp(APIItem):
    """Represents a DPI App configuration."""

    raw: TypedDPIRestrictionApp

    @property
    def id(self) -> str:
        """DPI app ID."""
        return self.raw["_id"]

    @property
    def apps(self) -> list[str]:
        """List of apps."""
        return self.raw["apps"]

    @property
    def blocked(self) -> bool:
        """Is blocked."""
        return self.raw["blocked"]

    @property
    def cats(self) -> list[str]:
        """Categories."""
        return self.raw["cats"]

    @property
    def enabled(self) -> bool:
        """Is enabled."""
        return self.raw["enabled"]

    @property
    def log(self) -> bool:
        """Is logging enabled."""
        return self.raw["log"]

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["site_id"]
